from abc import ABCMeta, abstractmethod

from series.logging import Logging


class PlayerInterface(Logging, metaclass=ABCMeta):

    def __init__(self, player):
        self._player = player

    @abstractmethod
    def stop(self):
        ...

    @abstractmethod
    def pause(self):
        ...

    @abstractmethod
    def osd(self, level):
        ...

    @abstractmethod
    def osd_level(self):
        ...

    def __getattr__(self, name):
        return getattr(self._player, name)

    def __setattr__(self, name, value):
        if name == '_player':
            super().__setattr__(name, value)
        else:
            setattr(self._player, name, value)

    @property
    def running(self):
        return self.time_pos is not None

__all__ = ['PlayerInterface']



"""add episode column removed

Revision ID: 3e0007af916
Revises: None
Create Date: 2014-01-09 04:28:09.160122

"""

# revision identifiers, used by Alembic.
revision = '3e0007af916'
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('episode', sa.Column('removed', sa.Boolean,
                                       server_default='0'))


def downgrade():
    op.drop_column('episode', 'removed')



import threading
import operator

import flask

from golgi.config import configurable

from series.rest_api import RestApi as RestApiBase, route_decorator

json = lambda: flask.request.json


def error(msg, status=404):
    return dict(error=msg), status


@configurable(library=['rest_api_host', 'rest_api_port'])
class RestApi(RestApiBase):

    routes, route = route_decorator()

    def __init__(self, library, player):
        RestApiBase.__init__(self, 'library')
        self._library = library
        self._player = player

    @route('/series')
    def series(self):
        series = self._library.all_series
        info = [seri.info for seri in series]
        return sorted(info, key=operator.itemgetter('name'))

    @route('/series/<name>')
    def series_data(self, name):
        series = self._library.series(name).info
        seasons = [s.info for s in self._library.seasons(series=name)]
        return dict(
            series=series,
            seasons=sorted(seasons, key=operator.itemgetter('number')),
        )

    @route('/series/<name>/<season>')
    @route('/series/<name>/season/<season>')
    def season_data(self, name, season):
        episodes = [e.info for e in self._library.episodes(series=name,
                                                           season=season)]
        series = self._library.series(name).info
        season = self._library.season(series=name, number=season).info
        return dict(
            series=series,
            season=season,
            episodes=sorted(episodes, key=operator.itemgetter('episode')),
        )

    @route('/series/<name>/<season>/recently_watched/<int:days>')
    @route('/series/<name>/season/<season>/recently_watched/<int:days>')
    def season_recent(self, days):
        seasons = self._library.recently_watched_seasons(days=days)
        return [season.info for season in seasons]

    @route('/episode/new')
    def new_episodes(self):
        episodes = self._library.new_episodes
        return [epi.info for epi in episodes]

    @route('/series/<series>/season/<season>/episode/<int:episode>',
           methods=['POST'])
    def create_episode_old(self, **kwargs):
        return self._library.create_episode(**kwargs).info

    @route('/series/<series>/seasons/<season>/episodes',
           methods=['POST'])
    def create_episode(self, series, season):
        if 'episode' in json():
            return self.create_episode_old(series=series, season=season,
                                           episode=json()['episode'])
        else:
            return error('No episode specified', 400)

    @route('/series/<series>/season/<season>/episode/<int:episode>',
           methods=['GET'])
    def episode(self, **data):
        result = self._library.episode(data['series'], data['season'],
                                       data['episode'])
        return result.info if result else error('No such episode')

    @route('/series/<series>/season/<season>/episode/<int:episode>',
           methods=['PUT'])
    @route('/series/<series>/seasons/<season>/episodes/<int:episode>',
           methods=['PUT'])
    def alter_episode(self, **kwargs):
        result = self._library.alter_episode(data=json(), **kwargs)
        return result.info if result else error('No such episode')

    @route('/series/<series>/season/<season>/episode/<int:episode>',
           methods=['DELETE'])
    def delete_episode(self, **kwargs):
        result = self._library.delete_episode(data=json(), **kwargs)
        return 'Success.' if result else error('No such episode')

    @route('/series/<series>/season/<season>', methods=['PUT'])
    def alter_season(self, **kwargs):
        result = self._library.alter_season(data=json(), **kwargs)
        return result.info if result else error('No such season')

    @route('/series/<series>', methods=['PUT'])
    def alter_series(self, **kwargs):
        result = self._library.alter_series(data=json(), **kwargs)
        return result.info if result else error('No such series')

    @route('/episode/recently_watched/<int:days>')
    def recently_watched_episodes(self, days):
        episodes = self._library.recently_watched_episodes(days=days)
        return [epi.info for epi in episodes]

    @route('/series/<series>/season/<season>/episode/<int:episode>/next')
    def next_episode(self, series, season, episode):
        info = None
        episode = self._library.episode(series, season, episode)
        if episode:
            _next = self._library.next_episode(episode)
            if _next:
                info = _next.info
        return dict(next=info)

    @route('/series/<series>/season/<season>/episode/<int:episode>/previous')
    def previous_episode(self, series, season, episode):
        info = None
        episode = self._library.episode(series, season, episode)
        if episode:
            _previous = self._library.previous_episode(episode)
            if _previous:
                info = _previous.info
        return dict(previous=info)

    @route('/movie')
    def movie(self):
        movies = self._library.all_movies
        info = [movie.info for movie in movies]
        return sorted(info, key=operator.itemgetter('title'))

    @route('/movie/<name>')
    def movie_data(self, name):
        movie = self._library.movie(name)
        if movie is not None:
            return movie.info

    @route('/movie/<title>/watch', methods=['PUT'])
    def watch_movie(self, title):
        movie = self._library.movie(title)
        if movie:
            self._player.target = movie
            self._player.start()
            return 'success'
        else:
            return error('Movie not found!')

    @route('/series/<series>/<int:season>/<int:episode>/watch',
           methods=['PUT'])
    @route('/series/<series>/season/<season>/episode/<int:episode>/watch',
           methods=['PUT'])
    def watch_episode(self, series, season, episode):
        epi = self._library.episode(series, season, episode)
        if epi:
            self._player.target = epi
            try:
                self._player.start()
            except OSError as e:
                return error(str(e))
            else:
                return 'success'
        else:
            return error('Episode not found')

    @route('/player/pause', methods=['PUT'])
    def player_toggle_pause(self):
        self._player.toggle_pause()

    @route('/player/seek/forward/<float:value>', methods=['PUT'])
    @route('/player/seek/forward/<int:value>', methods=['PUT'])
    def player_seek_forward(self, value):
        self._player.seek(value)

    @route('/player/seek/backward/<float:value>', methods=['PUT'])
    @route('/player/seek/backward/<int:value>', methods=['PUT'])
    def player_seek_backward(self, value):
        self._player.seek(-value)

    @route('/player/seek/<float:value>', methods=['PUT'])
    def player_seek_ratio(self, value):
        self._player.seek_to_ratio(value)

    @route('/player/subdelay/inc/<float:value>', methods=['PUT'])
    @route('/player/subdelay/inc/<int:value>', methods=['PUT'])
    def player_sub_delay_inc(self, value):
        self._player.sub_delay(value)

    @route('/player/subdelay/dec/<float:value>', methods=['PUT'])
    @route('/player/subdelay/dec/<int:value>', methods=['PUT'])
    def player_sub_delay_dec(self, value):
        self._player.sub_delay(-value)

    @route('/player/stop', methods=['PUT'])
    def player_stop(self):
        self._player.stop()

    @route('/player/volume/inc/<int:value>', methods=['PUT'])
    def player_volume_inc(self, value):
        self._player.change_volume(value)

    @route('/player/volume/dec/<int:value>', methods=['PUT'])
    def player_volume_dec(self, value):
        self._player.change_volume(-value)

    @route('/player/current')
    def currently_playing(self):
        item = self._player.target
        return item.info if item else None

    @route('/player/toggle_info', methods=['PUT'])
    def player_toggle_info(self):
        return self._player.toggle_info()

    @route('/collection/<_id>/new', methods=['PUT'])
    def scan(self, _id):
        if _id == 'all':
            target = self._library.scan
        elif _id == 'episode':
            target = self._library.scan_episodes
        elif _id == 'movie':
            target = self._library.scan_movies
        else:
            target = None
        if target is not None:
            worker = threading.Thread(target=target)
            worker.start()
            return 'Scanning initiated.'
        return 'No such collection.'

    @route('/clean', methods=['PUT'])
    def clean(self):
        worker = threading.Thread(target=self._library.clean)
        worker.start()
        return 'Cleaning initiated.'

__all__ = ['RestApi']



from golgi.config import configurable
from series.logging import Logging
from tek.tools import find

from tek_utils.sharehoster.torrent import torrent_cacher_client
from tek_utils.sharehoster.errors import ShareHosterError

from amino import List

from series.get.handler import ReleaseHandler, R, L
from series.get.errors import SeriesException
from series.condition import LambdaCondition, HandlerCondition, DynOr
from series.get.model.release import ReleaseMonitor
from series.get.model.link import TorrentLink


class CachableTorrent(HandlerCondition):

    def __init__(self, torrent: TorrentLink) -> None:
        self.torrent = torrent

    @property
    def cond(self):
        return ~(L('caching') | L('valid') | L('dead'))

    def ev(self, release: ReleaseMonitor):
        return self.cond.ev(self.torrent)

    def describe(self, item, target):
        d = self.cond.describe(self.torrent, target)
        m = self.torrent.url
        return '{} -> {}'.format(m, d)


class CachableTorrents(DynOr):

    @property
    def _sub_type(self):
        return CachableTorrent

    def _dyn_subs(self, item: ReleaseMonitor):
        return List.wrap(item.torrent_links)

    def describe(self, item, target):
        d = super().describe(item, target)
        return 'cachable torrents => {}'.format(d)


@configurable(torrent=['cacher'])
class TorrentHandler(ReleaseHandler):

    def __init__(self, releases, *a, **kw):
        self._pending = []
        super().__init__(releases, 5, 'torrent handler', **kw)
        self._client = torrent_cacher_client()

    def _check(self):
        super()._check()
        self._check_pending()
        self._check_error()

    def _handle(self, monitor):
        for torrent in monitor.cachable_torrents:
            msg = 'Requesting torrent download for {}…'.format(monitor.release)
            self.log.info(msg)
            self._pending.append([monitor, torrent])
            try:
                torrent.request()
            except ShareHosterError as e:
                msg = 'Error requesting torrent for {}: {}'
                self.log.error(msg.format(monitor.release, e))

    def _check_pending(self):
        done = []
        for monitor, torrent in self._pending:
            try:
                if torrent.cached:
                    self._releases.torrent_cached(torrent.record.id)
                    done.append([monitor, torrent])
                    msg = 'Flagging torrent as cached: {}'
                    self.log.info(msg.format(monitor.release))
            except ShareHosterError as e:
                msg = 'Error checking torrent status for {}: {}'
                self.log.error(msg.format(monitor.release, e))
        for item in done:
            self._pending.remove(item)

    def _check_error(self):
        errors = [t.get('id', 0) for t in self._client.transfers
                  if t.get('status') == 'ERROR']
        if errors:
            self.log.info('Canceling erroneous torrents.')
            self._client.cancel_transfers(errors)
            self._client.clean_transfers()

    @property
    def _conditions(self):
        return ~(
            R('downloaded') |
            LambdaCondition('in progress', self._requesting)
        ) & CachableTorrents()

    def _qualify(self, monitor):
        try:
            return super()._qualify(monitor)
        except ShareHosterError as e:
            msg = 'Error checking torrent status for {}: {}'
            self.log.error(msg.format(monitor.release, e))

    def _requesting(self, monitor) -> bool:
        return find(lambda item: item[0] == monitor, self._pending) is not None

    def _sanity_check(self):
        self._check_cacher_config()
        self._check_service_accessible()

    def _check_cacher_config(self):
        if not self._cacher:
            raise SeriesException('No torrent cacher configured!')

    def _check_service_accessible(self):
        cacher = torrent_cacher_client()
        if not cacher.account_info:
            raise SeriesException(
                'Couldn\'t access torrent cacher \'{}\'!'.format(self._cacher))

__all__ = ['TorrentHandler']



import abc
from typing import Callable, Any
from functools import wraps

from amino import List, Right, Left
from amino.util.string import snake_case

from series.logging import Logging


def exclusive(f: Callable[..., Any]):
    @wraps(f)
    def wrapper(self: 'DbFacade', *a, **kw):
        with self.lock:
            return f(self, *a, **kw)
    return wrapper


def commit(f: Callable[..., Any]):
    @wraps(f)
    def wrapper(self: 'DbFacade', *a, **kw):
        value = f(self, *a, **kw)
        self._commit()
        return value
    return exclusive(wrapper)


class DbFacade(Logging, metaclass=abc.ABCMeta):

    @abc.abstractproperty
    def main_type(self) -> type:
        ...

    def __init__(self, db):
        self._db = db

    @property
    def lock(self):
        return self._db.lock

    @exclusive
    def _commit(self):
        self._db.commit()
        self.log.debug('Committed transactions to db.')

    @property
    def all_query(self):
        return self._db.query(self.main_type).order_by(self.main_type.id)

    def filter_by(self, **filters):
        return self.all_query.filter_by(**filters)

    @property  # type: ignore
    @exclusive
    def all(self):
        return List.wrap(self.all_query.all())

    @exclusive
    def find_by_id(self, id):
        return self.filter_by(id=id).first()

    @property
    def name(self):
        return snake_case(str(self.main_type))

    def by_id(self, id):
        res = self.find_by_id(id)
        return Right(res) if res else Left(
            'No {} with id {}'.format(self.name, id))

__all__ = ('DbFacade', 'exclusive', 'commit')
